/* written by John Doug Reynolds, April 2002 */

#include "rgsregions.h"
#include <ShapeScanner.h>
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <CalServer.h>
#include <DataFilter.h>
#include <ASCregion.h>
#include <cstdio>

namespace {
   using namespace rgsPixelsLib;
   using std::string; using std::vector; using std::list;

   class RegionTrait {
   public:
      virtual ~RegionTrait() {};
      virtual bool test(const Block*) const = 0;
      virtual void addTo(Block*) const = 0;
   };

   template<class T>
      class RegTraitInst : public RegionTrait {
      public:
	 RegTraitInst(const string &name, T val, const string &comment)
	    : _name(name), _comment(comment), _value(val) {}
	 T value() const { return _value; }
	 bool test(const Block*) const;
	 void addTo(Block *block) const
	    { *(block->addAttribute(_name,_comment)) = _value; }
      private:
	 string _name, _comment;
	 T _value;
      };

   class RegionManager {
   public:
      ~RegionManager();
      RegionManager( const DataSet *evset, bool protect,
		     int spatialres, int orderres, float xpsfbelow,
		     float xpsfabove, float xpsfexcl, float pdistincl );

      void exposed( struct DataSet*, unsigned sourceid ) const;

      void image( struct DataSet*, unsigned sourceid, float xdspSrc ) const;

      void banana( struct DataSet*, unsigned sourceid, unsigned order,
		   float dispSrc, bool background = false ) const;

      void background( struct DataSet*, const vector<unsigned> &sourceid,
		       const vector<float> &xdspSrc ) const;

   private:
      class Node {
      public:
	 int ccd, node;
	 Node( const Array*, unsigned spatialres, unsigned orderres );
	 FShape* exposed() const;
	 FShape* image( float coverageBelow, float coverageAbove,
			float rgsXdspAngle, struct CanonicalCrossPsf* ) const;
	 FShape* banana( float coverage, float alpha, float linespacing,
			 unsigned order, bool background,
			 struct CanonicalRedist* ) const;
	 FShape* calib( const Spectrum& ) const;
      private:
	 struct VertexPair {
	    float x, y0, y1;
	    VertexPair( float _x, float _y0, float _y1 )
	       : x(_x), y0(_y0), y1(_y1) {}
	 };

	 struct GridSpan _betaGrid, _xdspGrid;
	 vector<float> _betaBounds, _xdspBounds, _iBeta, _eBeta;
	 CalReal64Vector _xdsp;

	 static FPolygon* polygon( const list<VertexPair>& );
      };

      bool _protect;
      struct RegTraitInst<int> _spatialres, _orderres;
      struct RegTraitInst<float> _xpsfbelow, _xpsfabove, _xpsfexcl, _pdistincl;
      struct GridSpan _betaGrid, _xdspGrid;
      struct CanonicalCrossPsf *_crossPSF;
      struct CanonicalRedist *_redist;
      float _alpha0, _linespacing;
      string _instrument;
      list<Node*> _nodes;

      static string toStr( unsigned );
      static bool test( struct DataSet*, const string &extname,
			const list<const RegionTrait*>& );
      void install( struct DataSet*, const GeoRegion&,
		    const string &extname, const string &ylabel,
		    const list<const RegionTrait*>&,
		    const Grid &x, const Grid &y ) const;
   };
}

static
void select( vector<bool> &flags, struct Table *table,
	     const string &flagName, const string &expression, bool setflags );

void rgsregions()
{
   string dsname	= stringParameter("srclist");
   string evname	= stringParameter("evlist");
   unsigned maxorder	= intParameter("maxorder");
   string procsrcsexpr	= stringParameter("procsrcsexpr");
   string exclsrcsexpr	= stringParameter("exclsrcsexpr");
   bool setflags	= booleanParameter("setflags");
   int spatialres	= intParameter("spatialres");
   int orderres		= intParameter("orderres");
   float xpsfbelow	= realParameter("xpsfbelow");
   float xpsfabove	= realParameter("xpsfabove");
   float xpsfexcl	= realParameter("xpsfexcl");
   float pdistincl	= realParameter("pdistincl");
   bool protect		= booleanParameter("protect");

   if ( xpsfbelow > 0 && xpsfbelow < 1 )
      Sas::warning("fractionalCoverage",
		   "the value provided for xpsfbelow suggests that a fraction "
		   "was given where a percentage is required");
   if ( xpsfabove > 0 && xpsfabove < 1 )
      Sas::warning("fractionalCoverage",
		   "the value provided for xpsfabove suggests that a fraction "
		   "was given where a percentage is required");
   if ( xpsfexcl > 0 && xpsfexcl < 1 )
      Sas::warning("fractionalCoverage",
		   "the value provided for xpsfexcl suggests that a fraction "
		   "was given where a percentage is required");
   if ( pdistincl > 0 && pdistincl < 1 )
      Sas::warning("fractionalCoverage",
		   "the value provided for pdistincl suggests that a fraction "
		   "was given where a percentage is required");

   if ( evname == "" ) {
      evname = dsname;

      string::size_type end = evname.rfind('.');
      if ( end == string::npos ) end = evname.size();

      if ( end >= 10 )
	 evname.replace( end-10, 6, "EVENLI" );
      else
	 evname.replace( 0, end, "evlist" );
   }

   struct DataSet *evset = dataSetServer->open( evname, Dal::Read );
   enum Xmm::DataMode datamode
      = Xmm::toDataMode( evset->attribute("DATAMODE")->asString() );
   bool modeIsHTR = (datamode != Xmm::DATA_SPECTROSCOPY);
   struct RegionManager manager( evset, protect, spatialres, orderres,
				 xpsfbelow, xpsfabove, xpsfexcl, pdistincl );
   dataSetServer->close(evset);

   struct DataSet *ds = dataSetServer->open( dsname, Dal::Modify );
   {
      const float toRad = M_PI / 10800;
      unsigned i;

      struct Table *stab = ds->table("SRCLIST");
      const DalVecI index( stab->column("INDEX") );
      const DalVecE rgsDisp( stab->column("DELTA_DISP") );
      const DalVecE rgsXdsp( stab->column("DELTA_XDSP") );

      vector<bool> proc, excl;
      if ( modeIsHTR ) exclsrcsexpr = "false";
      select( excl, stab, "BKG_EXCLUDE", exclsrcsexpr, setflags );
      select( proc, stab, "PROCESS", procsrcsexpr, setflags );

      if ( !modeIsHTR ) {
	 vector<unsigned> sourceid;
	 vector<float> xdspSrc;
	 for ( i = 0; i < excl.size(); ++i )
	    if ( excl[i] ) {
	       sourceid.push_back( index[i] );
	       xdspSrc.push_back( toRad * rgsXdsp[i] );
	    }
	 manager.background( ds, sourceid, xdspSrc );
      }

      for ( i = 0; i < proc.size(); ++i )
	 if ( proc[i] ) {
	    if ( modeIsHTR )
	       manager.exposed( ds, index[i] );
	    else
	       manager.image( ds, index[i], toRad * rgsXdsp[i] );

	    for ( unsigned order = 1; order <= maxorder; ++order ) {
	       manager.banana( ds, index[i], order, toRad * rgsDisp[i] );
	       if ( modeIsHTR )
		  manager.banana( ds, index[i], order,
				  toRad * rgsDisp[i], true /*background*/ );
	    }
	 }
   }
   dataSetServer->close(ds);
}

void select( vector<bool> &flags, struct Table *table,
	     const string &flagName, const string &expression, bool setflags )
{
   flags.resize( table->rows() );
   DalVecL flagColumn( table->column(flagName) );
   unsigned i;

   if ( expression == "" ) {
      for ( i = 0; i < table->rows(); ++i ) flags[i] = flagColumn[i];
      return;
   }

   vector<DataFilter::RowRange> ranges;
   DataFilter(expression).doSelect(*table,ranges);

   vector<DataFilter::RowRange>::const_iterator itr;
   for ( itr = ranges.begin(); itr != ranges.end(); ++itr )
      for ( i = itr->startRow; i <= itr->stopRow; ++i ) flags[i] = true;

   if ( setflags )
      for ( i = 0; i < table->rows(); ++i ) flagColumn[i] = flags[i];
}

/* --- RegTraitInst specializations ---------------------------------------- */

bool RegTraitInst<int>::test(const Block *block) const
{
   return block->hasAttribute(_name) &&
      block->attribute(_name)->asInt() == _value;
}
bool RegTraitInst<float>::test(const Block *block) const
{
   return block->hasAttribute(_name) &&
      block->attribute(_name)->asReal() == _value;
}
bool RegTraitInst<string>::test(const Block *block) const
{
   return block->hasAttribute(_name) &&
      block->attribute(_name)->asString() == _value;
}

/* --- RegionManager::Node methods ----------------------------------------- */

static const float xdspMargin = 1e-4;

static
void init( vector<float> &vertex, const GridSpan &grid, unsigned nAcross )
{
   vertex.resize( nAcross );

   float stride = grid.stride * grid.length() / (nAcross - 1);

   unsigned i = 0;
   for ( float beta = grid.minval(); i < nAcross; beta += stride )
      vertex[i++] = beta;
}

RegionManager::Node::
Node( const Array *emap, unsigned spatialres, unsigned orderres )
   : ccd( emap->attribute("CCDID")->asInt() ),
     node( emap->attribute("CCDNODE")->asInt() ),
     _betaGrid( gridspan(emap,1) ),
     _xdspGrid( gridspan(emap,2) ),
     _betaBounds(4), _xdspBounds(4),
     _xdsp( 0, _xdspGrid.length() + int(2*xdspMargin/_xdspGrid.stride) )
{
   _betaBounds[0] = _betaBounds[3] = _betaGrid.minval();
   _betaBounds[1] = _betaBounds[2] = _betaGrid.maxval();
   _xdspBounds[0] = _xdspBounds[1] = _xdspGrid.minval();
   _xdspBounds[2] = _xdspBounds[3] = _xdspGrid.maxval();

   init( _iBeta, _betaGrid, spatialres );
   init( _eBeta, _betaGrid, orderres );

   int c = _xdspGrid.min - int(xdspMargin/_xdspGrid.stride);
   for ( unsigned i = 0; i < _xdsp.size(); _xdsp[i++] = _xdspGrid[c++] );
}

static
void narrow( const Spectrum &dist,
	     unsigned anchor, float coverageBelow, float coverageAbove,
	     unsigned &begin, unsigned &end )
{
   begin = 0;
   if ( (end = dist.size()) == 0 ) return;
   if ( anchor >= end ) anchor = end - 1;

   unsigned i;
   float area;

   for ( area = 0, i = begin; i <= anchor; area += dist[i++] );
   area *= 1 - coverageBelow; // excess area below
   while ( begin < anchor && dist[begin] <= area ) area -= dist[begin++];

   for ( area = 0, i = anchor; i < end; area += dist[i++] );
   area *= 1 - coverageAbove; // excess area above
   while ( --end > anchor && dist[end] <= area ) area -= dist[end];
   ++end;
}

FShape* RegionManager::Node::image( float coverageBelow, float coverageAbove,
				    float xdspSrc,
				    struct CanonicalCrossPsf *xpsf ) const
{
   int cell0 = _xdspGrid.cell( _xdsp[0] );
   int anchor = _xdspGrid.cell( -xdspSrc ) - cell0;
   float halfstep = _xdspGrid.stride / 2;

   if ( anchor < 0 || anchor >= (int)_xdsp.size() ) return 0;

   bool manifest = false;
   list<VertexPair> vertices;
   for ( unsigned i = 0; i < _iBeta.size(); ++i ) {
      unsigned begin, end;
      narrow( xpsf->probability(0,_iBeta[i],xdspSrc,_xdsp,_xdspGrid.stride),
	      anchor, coverageBelow, coverageAbove, begin, end );

      minimize(   end, 1u+(_xdspGrid.max-cell0) ); minimize( begin, end );
      maximize( begin, 0u+(_xdspGrid.min-cell0) ); maximize( end, begin );
      if ( end > begin ) manifest = true;

      float xdsp0 = _xdsp[begin] - halfstep, xdsp1 = _xdsp[end-1] + halfstep;
      vertices.push_back( VertexPair(_iBeta[i],xdsp0,xdsp1) );
   }

   return manifest ? polygon(vertices) : 0;
}

static
void narrow( const Spectrum &dist,
	     float coverage, unsigned &begin, unsigned &end )
{
   begin = 0;
   end = dist.size();

   float area = 0;
   Spectrum::const_iterator iter;
   for ( iter = dist.begin(); iter != dist.end(); area += *iter++ );
   area *= 1 - coverage; // excess area

   bool left = false;
   while ( end > begin ) {
      float v0 = dist[begin], v1 = dist[end-1];
      if ( v0 > area && v1 > area ) return;
      left = v0 >= v1 ? v0 == v1 ? !left : false : true;
      area -= dist[ left ? begin++ : --end ];
   }
}

FShape*
RegionManager::Node::banana( float coverage, float alpha, float linespacing,
			     unsigned order, bool background,
			     struct CanonicalRedist *redist ) const
{
   const float siEdge = 1837.95, delta = 1.0; // eV

   float cosAlpha = cos( alpha );
   float multiplier = 12398. * -int(order) / linespacing;
   float siBetaEdge = acos( multiplier/siEdge + cosAlpha );
   float siBetaBelow = acos( multiplier/(siEdge+delta) + cosAlpha );
   float siBetaAbove = acos( multiplier/(siEdge-delta) + cosAlpha );

   list<float> beta( _eBeta.begin(), _eBeta.end() );

   list<float>::iterator itr, prv;
   for ( itr = beta.begin(); itr != beta.end() && *itr < siBetaEdge; ++itr );
   if ( itr != beta.begin() && itr != beta.end() ) {
      if ( siBetaBelow > *--(prv=itr) ) beta.insert(itr,siBetaBelow);
      if ( siBetaAbove < *itr ) beta.insert(itr,siBetaAbove);
   }

   list<VertexPair> vertices;
   for ( itr = beta.begin(); itr != beta.end(); ++itr ) {
      unsigned begin, end;
      float energy = multiplier / (cos(*itr) - cosAlpha);
      narrow( redist->spectrum(energy), coverage, begin, end );
      if ( background ) {
	 unsigned length = end - begin;
	 begin = end;
	 end = begin + length;
      }
      if ( end > begin ) vertices.push_back( VertexPair(*itr,begin,end) );
   }

   return polygon(vertices);
}

FShape* RegionManager::Node::calib( const Spectrum &dist ) const
{
   unsigned begin, end;
   narrow( dist, .9, begin, end );

   list<VertexPair> vertices;
   vertices.push_back( VertexPair(_eBeta[0],begin,end) );
   vertices.push_back( VertexPair(_eBeta[_eBeta.size()-1],begin,end) );

   return polygon(vertices);
}

FShape* RegionManager::Node::exposed() const
{
   return new FPolygon(_betaBounds,_xdspBounds,_xdspBounds.size());
}

FPolygon* RegionManager::Node::polygon( const list<VertexPair> &pairs )
{
   if ( pairs.size() < 2 ) return 0;

   unsigned size = 2 * pairs.size();

   vector<float> x(size), y(size);

   unsigned i = 0, j = size-1;
   list<VertexPair>::const_iterator itr;
   for ( itr = pairs.begin(); itr != pairs.end(); ++itr ) {
      x[i] = x[j] = itr->x;
      y[i++] = itr->y0;
      y[j--] = itr->y1;
   }

   return new FPolygon(x,y,size);
}

/* --- RegionManager methods ----------------------------------------------- */

RegionManager::~RegionManager()
{
   list<Node*>::iterator itr;
   for ( itr = _nodes.begin(); itr != _nodes.end(); ++itr ) delete *itr;
   calServer.releaseAtom(_crossPSF);
   calServer.releaseAtom(_redist);
}

RegionManager::RegionManager(const DataSet *evset, bool protect,
			     int spatialres, int orderres, float xpsfbelow,
			     float xpsfabove, float xpsfexcl, float pdistincl)
   :
   _protect(protect),
   _spatialres("SPACING",spatialres,"number of vertices across each node"),
   _orderres("SPACING",orderres,"number of vertices across each node"),
   _xpsfbelow("COVBELOW",xpsfbelow,"percent of lower cross-disp PSF included"),
   _xpsfabove("COVABOVE",xpsfabove,"percent of upper cross-disp PSF included"),
   _xpsfexcl("COVERAGE",xpsfexcl,"percent of cross-disp PSF excluded"),
   _pdistincl("COVERAGE",pdistincl,"percent of PI distribution included")
{
   const Table *etab = evset->table("EVENTS");
   _betaGrid = gridspan( etab->column("BETA_CHANNEL") );
   _xdspGrid = gridspan( etab->column("XDSP_CHANNEL") );

   _instrument = evset->attribute("INSTRUME")->asString();
   calServer.setInstrument( Xmm::toInstrument(_instrument) );
   calServer.state()->ccd()->set(1); // any CCD is fine
   _crossPSF = 0; _crossPSF = calServer.getAtom(_crossPSF);
   _redist = 0; _redist = calServer.getAtom(_redist);
   _redist->failSafe(true);

   struct MiscDataServer *misc = 0; misc = calServer.getAtom(misc);
   _alpha0 = (misc->get("INCIDENCE_ANGLE")
	      + misc->get("LINCOORD:LINCOORD%G_RY")) * M_PI/180;
   _linespacing = misc->get("GRAT_LINE_DENS");
   calServer.releaseAtom(misc);

   for ( unsigned ccd = 1; ccd <= 9; ++ccd )
      for ( unsigned node = 0; node < 2; ++node ) {
	 string extname = string("EXPMAP")+char('0'+node)+char('0'+ccd);
	 if ( !evset->hasBlock(extname) ) continue;
	 const Array *emap = evset->array(extname);
	 if ( emap->size()[0] && emap->size()[1] )
	    _nodes.push_back( new Node(emap,spatialres,orderres) );
      }

   if ( _nodes.size() == 0 )
      Sas::error("noExposureMaps","the event list contains no exposure maps");
}

void RegionManager::exposed( struct DataSet *ds, unsigned sourceid ) const
{
   string extname = _instrument + "_SRC" + toStr(sourceid) + "_SPATIAL";

   list<const RegionTrait*> traits;
   if ( test(ds,extname,traits) ) return;

   struct GeoRegion region;

   list<Node*>::const_iterator itr;
   for ( itr = _nodes.begin(); itr != _nodes.end(); ++itr )
      region.push_back( Component(1,(*itr)->exposed()) );

   install( ds, region, extname, "XDSP_CORR", traits, _betaGrid, _xdspGrid );
}

void RegionManager::image( struct DataSet *ds,
			   unsigned sourceid, float xdspSrc ) const
{
   string extname = _instrument + "_SRC" + toStr(sourceid) + "_SPATIAL";

   list<const RegionTrait*> traits;
   traits.push_back(&_xpsfbelow);
   traits.push_back(&_xpsfabove);
   traits.push_back(&_spatialres);
   if ( test(ds,extname,traits) ) return;

   struct GeoRegion region;
   float covBelow = _xpsfbelow.value() / 100;
   float covAbove = _xpsfabove.value() / 100;

   list<Node*>::const_iterator itr;
   for ( itr = _nodes.begin(); itr != _nodes.end(); ++itr ) {
      FShape* shape = (*itr)->image(covBelow,covAbove,xdspSrc,_crossPSF);
      if ( shape ) region.push_back( Component(1,shape) );
   }

   install( ds, region, extname, "XDSP_CORR", traits, _betaGrid, _xdspGrid );
}

void RegionManager::banana( struct DataSet *ds,
			    unsigned sourceid, unsigned order,
			    float dispSrc, bool background ) const
{
   string extname = _instrument + "_SRC" + toStr(sourceid) + "_ORD";
   if ( background )
      extname += toStr(order) + "_BKG";
   else
      extname += "ER_" + toStr(order);

   list<const RegionTrait*> traits;
   traits.push_back(&_pdistincl);
   traits.push_back(&_orderres);
   if ( test(ds,extname,traits) ) return;

   struct GeoRegion region;
   float coverage = _pdistincl.value() / 100;
   float alpha = _alpha0 + dispSrc;

   list<Node*>::const_iterator itr;
   for ( itr = _nodes.begin(); itr != _nodes.end(); ++itr ) {
      unsigned effectiveOrder = order;
      if ( background )
	 switch ( order ) {
	 case 1: if ( (*itr)->ccd < 4 ) ++effectiveOrder; break;
	 }

      FShape *shape=(*itr)->banana(coverage,alpha,_linespacing,
				   effectiveOrder,background,_redist);
      if ( shape == 0 ) continue;

      float lambda = 0;
      switch ( (*itr)->node + 2 * (*itr)->ccd - 2 ) {
      case 1: case 2: case 11: case 12: lambda =  8.338; break;
      case 3: case 4: case 13: case 14: lambda = 18.319; break;
      }

      Component node( 1, shape );
      if ( lambda ) {
	 node.push_back( (*itr)->calib(_redist->spectrum(12398/lambda)) );
	 node.back()->exclude();
      }
      region.push_back( node );
   }

   install( ds, region, extname, "PI", traits, _betaGrid, Grid() );
}

void RegionManager::background( struct DataSet *ds,
				const vector<unsigned> &sourceid,
				const vector<float> &xdspSrc ) const
{
   string srclist;
   if ( sourceid.size() ) {
      vector<unsigned>::const_iterator i = sourceid.begin();
      for ( srclist = toStr(*i++); i != sourceid.end(); ++i )
	 srclist += " " + toStr(*i);
   }
   else srclist = "none";
   struct RegTraitInst<string> exclsrcs("EXCLSRCS",srclist,"excluded sources");

   string extname = _instrument + "_BACKGROUND";

   list<const RegionTrait*> traits;
   traits.push_back(&exclsrcs);
   traits.push_back(&_xpsfexcl);
   traits.push_back(&_spatialres);
   if ( test(ds,extname,traits) ) return;

   struct GeoRegion region;
   float coverage = _xpsfexcl.value() / 100;

   list<Node*>::const_iterator itr;
   for ( itr = _nodes.begin(); itr != _nodes.end(); ++itr ) {
      Component node( 1, (*itr)->exposed() );

      vector<float>::const_iterator xdspSrcI;
      for ( xdspSrcI=xdspSrc.begin(); xdspSrcI != xdspSrc.end(); ++xdspSrcI ) {
	 FShape *shape = (*itr)->image(coverage,coverage,*xdspSrcI,_crossPSF);
	 if ( shape ) {
	    shape->exclude();
	    node.push_back(shape);
	 }
      }

      region.push_back( node );
   }

   install( ds, region, extname, "XDSP_CORR", traits, _betaGrid, _xdspGrid );
}

void RegionManager::install( struct DataSet *ds, const GeoRegion &region,
			     const string &extname, const string &ylabel,
			     const list<const RegionTrait*> &traits,
			     const Grid &xgrid, const Grid &ygrid ) const
{
   int col;
   unsigned count = 0;
   list<Span> spans;
   list<Span>::const_iterator itr;
   struct RegionScanner scanner( region, xgrid, ygrid );
   for ( scanner.reset(); scanner.next(col,spans); )
      for ( itr = spans.begin(); itr != spans.end(); ++itr )
	 count += itr->length();

   if ( count == 0 )
      errstr << "region " << extname << " encloses no pixels"
	     << Sas::warning("emptyRegion");

   struct Table *table = ds->addTable( extname, region.elements() );
   struct OutASCregionTable regionTable(table,"BETA_CORR",ylabel);
   regionTable << region;

   *(table->addAttribute("PROTECT","protected from replacement"))
      = _protect;

   *(table->addAttribute("PIX_ENCL","number of pixels enclosed"))
      = int(count);

   *(table->addAttribute("PIX_AREA","area of a pixel"))
      = xgrid.stride * ygrid.stride;

   list<const RegionTrait*>::const_iterator trait;
   for ( trait = traits.begin(); trait != traits.end(); ++trait )
      (*trait)->addTo(table);
}

bool RegionManager::test( struct DataSet *ds, const string &extname,
			  const list<const RegionTrait*> &traits )
{
   if ( ! ds->hasBlock(extname) ) {
      errstr << "adding region " << extname
	     << Sas::message(Msg::App,Msg::Sparse);
      return false;
   }

   const Block *block = ds->block(extname);
   list<const RegionTrait*>::const_iterator trait;
   for ( trait = traits.begin(); trait != traits.end(); ++trait )
      if ( ! (*trait)->test(block) )
	 if ( block->hasAttribute("PROTECT") &&
	      block->attribute("PROTECT")->asBool() == false ) {
	    errstr << "replacing region " << extname
		   << Sas::message(Msg::App,Msg::Sparse);
	    ds->deleteBlock(extname);
	    return false;
	 }
	 else {
	    errstr << "region " << extname << " is protected and therefore "
		   << "cannot be replaced with the specified region"
		   << Sas::warning("protectedRegion");
	    return true;
	 }

   return true;
}

string RegionManager::toStr( unsigned i )
{
   char buf[16];
   std::sprintf(buf,"%u",i);
   return buf;
}
